# Copyright 2023 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

IS_WEIGHTS = "is_weights"
IS_ACTIVATIONS = "is_activations"

WEIGHTS_QUANTIZATION_METHOD = 'weights_quantization_method'
WEIGHTS_N_BITS = 'weights_n_bits'
WEIGHTS_QUANTIZATION_PARAMS = 'weights_quantization_params'
ENABLE_WEIGHTS_QUANTIZATION = 'enable_weights_quantization'
WEIGHTS_CHANNELS_AXIS = 'weights_channels_axis'
WEIGHTS_PER_CHANNEL_THRESHOLD = 'weights_per_channel_threshold'
MIN_THRESHOLD = 'min_threshold'
ACTIVATION_QUANTIZATION_METHOD = 'activation_quantization_method'
ACTIVATION_N_BITS = 'activation_n_bits'
ACTIVATION_QUANTIZATION_PARAMS = 'activation_quantization_params'
ENABLE_ACTIVATION_QUANTIZATION = 'enable_activation_quantization'


